package exercises.ex_z4;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import static exercises.ex_z4.AuctionState.*;

public class Auctioneer implements Runnable {
    private static final DecimalFormat df = new DecimalFormat("#.00");

    private final String name;
    private final Queue<Bid> bids = new ConcurrentLinkedQueue<>();
    private final ArrayList<Item> items;
    private final Lock bidLock = new ReentrantLock();
    private final Condition bidCond = bidLock.newCondition();
    public final Lock stateLock = new ReentrantLock();
    public final Condition stateCond = stateLock.newCondition();

    private Bid maximumBid;
    private Thread countdownThread = null;
    private AuctionState auctionState = PRESENTING_ITEM;


    public Auctioneer(String name, ArrayList<Item> items) {
        this.name = name;
        this.items = items;
    }

    private void announce(String msg) {
        System.out.println("[Auctioneer " + name + "] " + msg);
    }

    private void setState(AuctionState newState) {
        stateLock.lock();
        try {
            auctionState = newState;
            stateCond.signalAll();
        } finally {
            stateLock.unlock();
        }
    }

    public void presentItem() throws InterruptedException {
        Item item = items.getFirst();
        announce("Now selling this new item: '" + item.name() + "'");
        announce("Get ready to bid...");
        Thread.sleep(2000);
        announce("Bidding open !");
        maximumBid = null;
        setState(BIDDING);
    }

    public void placeBid(Bid bid) {
        if (auctionState == BIDDING) {
            System.out.println(bid.bidder() + " bids $" + df.format(bid.amount()));
            bidLock.lock();
            try {
                bids.add(bid);
                bidCond.signalAll();
            } finally {
                bidLock.unlock();
            }
        }
    }

    private void countdown() {
        try {
            Thread.sleep(1000);
            System.out.println("Three");
            Thread.sleep(1000);
            System.out.println("Two");
            Thread.sleep(1000);
            System.out.println("One");
            Thread.sleep(1000);
            setState(SOLD);
            bidLock.lock();
            try {
                bidCond.signalAll();
            } finally {
                bidLock.unlock();
            }
        } catch (InterruptedException _) {}
    }

    public void startCountdown() {
        countdownThread = new Thread(this::countdown);
        countdownThread.start();
    }

    public Bid getMaximumBid() {
        return maximumBid;
    }

    public AuctionState getAuctionState() {
        return auctionState;
    }

    public void finalizeBidding() throws InterruptedException {
        if (maximumBid == null) {
            announce("Nobody wants this items ?!");
        } else {
            announce("And sold for $" + df.format(maximumBid.amount()) + " !");
            Item item = items.removeFirst();
            announce(maximumBid.bidder() + " is now the proud owner of " + item);
            maximumBid.bidder().addItem(item);
        }
        Thread.sleep(2000);
        if (items.isEmpty()) {
            setState(FINISHED);
        } else {
            setState(PRESENTING_ITEM);
        }
    }

    @Override
    public void run() {
        while (auctionState != FINISHED) {
            try {
                switch (auctionState) {
                    case PRESENTING_ITEM -> {
                        presentItem();
                    }
                    case BIDDING -> {
                        Bid current_bid = bids.poll();
                        if (current_bid == null) {
                            if (maximumBid != null) {
                                startCountdown();
                            }
                            bidLock.lock();
                            try {
                                bidCond.await();
                            } finally {
                                bidLock.unlock();
                            }
                            if (countdownThread != null) {
                                countdownThread.interrupt();
                            }
                        } else {
                            if (maximumBid == null || current_bid.amount() > maximumBid.amount()) {
                                maximumBid = current_bid;
                                announce("New best bid : " + df.format(maximumBid.amount()) + " by " + maximumBid.bidder());
                            }
                        }
                    }
                    case SOLD -> {
                        finalizeBidding();
                    }
                }

            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
        }
        announce("The auction is now finished ! Thank you for participating !");
    }
}
